#include <sys/wait.h>
#include <sys/types.h>
#include <signal.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/select.h>

#include <netinet/in.h>
#include <arpa/inet.h>


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <pthread.h>

static __thread unsigned int tlsData = 0;
static volatile char runtask1 = 1;
static volatile char task1ended = 0;

struct task1params {
  int bindPort;
  int bindFd;

  int remoteFd;
  pthread_t remoteTid;
  volatile char runListener;
};

/**
 *
 *
 *
 */
/*static void processReaper( int signal )
{
  int status, pid;

  fprintf( stderr, "%lu: i'm the process reaper\n", pthread_self() );
  if( signal != SIGCHLD ) {
    fprintf( stderr, "%lu: received something else that a SIGCHLD(%d)\n", pthread_self(), signal );
    return;
  }
  while( (pid = wait4(0, &status, WNOHANG, NULL)) > 0) {
    if( WIFSIGNALED(status) ) {
      fprintf( stderr, "%lu: child %d ended abnormaly(%d)\n", pthread_self(), pid, WEXITSTATUS(status) );
    }
  }  
  }*/


 /**
  *  This is the listener thread, it just checks that it has something to read
  *  by selecting for 1 second
  *
  */
void *remoteworker_thread_run( void *args ) 
{
  struct task1params *params;
  struct timeval timeout;
  fd_set read_fds;
  int n, retRead;
  char buffer[300];

  printf("%d(%lu)remoteworker_thread_run: starting\n", getpid(), pthread_self() );
  if( args == NULL ) {
    fprintf( stderr, "%s: no args, exiting\n", __func__ );
    return NULL;
  }
  params = (struct task1params *)args;

  // speaks a little to the client
  snprintf( buffer, sizeof(buffer), "toto\n" );
  if( write(params->remoteFd, buffer, strlen(buffer)) < 0 ) {
    fprintf( stderr, "%s: error in write(%s)\n", __func__, strerror(errno) );
    return NULL;    
  }

  while( params->runListener ) {
    FD_ZERO(&read_fds);
    FD_SET(params->remoteFd, &read_fds );
    timeout.tv_sec = 1;
    timeout.tv_usec = 0;
    n = select(params->remoteFd + 1, &read_fds, NULL, NULL, &timeout );
    if( n < 0 ) {
         fprintf( stderr, "%s: something failed while waiting(%s)\n", __func__, strerror(errno) );
	 continue;
    } else if( !n || !FD_ISSET(params->remoteFd, &read_fds) ) continue;
    
    retRead = read( params->remoteFd, buffer, sizeof(buffer)-1 );
    if( retRead < 0 ) {
      if( errno == EINTR) fprintf(stderr, "%s: interrupted while reading\n", __func__ );
      else {
	fprintf(stderr, "%s: error while reading, exiting....\n", __func__ );
	break;
      }
    }
    buffer[ retRead ] = '\0';
    printf("%s(%lu): readed %u bytes [%s]\n", __func__, pthread_self(), retRead, buffer );
  }
  return NULL;
}

/**
 *   Verifies that /sbin/true is really true(we never know)
 *
 *
 */
void task1_core( void )
{
  FILE *fd;
  char buffer[100];

  tlsData++;
  fd = popen("/bin/true", "r");
  if( fd == NULL ) {
    fprintf(stderr, "%s(%lu): popen failed(%s)\n", __func__, pthread_self(), strerror(errno) );
    sleep(10);
    return;
  }
  if( fgets( buffer, sizeof(buffer), fd) == NULL ) {
    if( pclose(fd) < 0 ) {
      fprintf(stderr, "%s(%lu): pclose failed(%s)\n", __func__, pthread_self(), strerror(errno) );
      sleep(10);
    } /*else {
      fprintf(stderr, "%s(%lu): fgets failed(%s)\n", __func__, pthread_self(), strerror(errno) );
      }*/
    return;
  }
  if( pclose(fd) < 0 ) {
    fprintf(stderr, "%s(%lu): pclose failed(%s)\n", __func__, pthread_self(), strerror(errno) );
    sleep(10);
    return;
  }
}

/**
 *
 *
 *
 */
void bindPort( struct task1params *params )
{
  struct sockaddr_in addr;

  struct sockaddr_in remoteAddr;
  int remoteAddrLen;

  if(!params) {
    fprintf(stderr, "%s: don't give me an empty params\n", __func__ );
    return;
  }
  
  if( (params->bindFd = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
    fprintf(stderr, "%s: unable to create a socket(%s)\n", __func__, strerror(errno) );
    return;    
  }
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = htonl(INADDR_ANY);
  addr.sin_port = htons(params->bindPort);

  if( bind(params->bindFd, (struct sockaddr *)&addr, sizeof(addr)) < 0 ) {
    fprintf( stderr, "%s: unable to bind(%s)\n", __func__, strerror(errno) );
    close(params->bindFd);
    return;
  }

  if( listen(params->bindFd, 10) < 0 ) {
    fprintf( stderr, "%s: unable to listen(%s)\n", __func__, strerror(errno) );
    close(params->bindFd);
    return;
  }
  if( (params->remoteFd = accept(params->bindFd, (struct sockaddr *)&remoteAddr, &remoteAddrLen)) < 0 ) {
    fprintf( stderr, "%s: error in accept(%s)\n", __func__, strerror(errno) );
    return;
  }

  // once somebody has been accepted launch a thread treating the request
  if( pthread_create( &params->remoteTid, NULL, remoteworker_thread_run, params) != 0 ) {
    fprintf(stderr, "%s: unable to create listener thread\n", __func__ );
    return;
  }
}

void unbindPort( struct task1params *params )
{
  close( params->remoteFd );
  close( params->bindFd );
}

/**
 *
 *
 *
 */
void *do_task1( void *p_args )
{
  struct task1params *params = (struct task1params *)p_args;

  printf("%d(%lu)do_task1: starting\n", getpid(), pthread_self() );
  if( p_args ) {
    bindPort( params );
  };
  do {
    sleep(5);
    task1_core();
    printf("%s(%lu): tlsData=%u\n", __func__, pthread_self(), tlsData );
  } while( runtask1 );
  
  if( p_args ) {
    unbindPort( params );
  };
  task1ended = 1;
  return NULL;
}


int main( int argc, char *argv[] )
{
  pthread_t task1, task2;
  int counter;
  struct task1params params, *task1args = NULL;

  /*  if( signal(SIGCHLD, processReaper) == SIG_ERR ) {
    fprintf( stderr, "%s(%lu): error setting SIGCHLD handler\n", __func__, pthread_self() );
    return 1;
    }*/
  
  if( argc > 2 ) task1args = &params;
  params.bindPort = 7899;
  if( pthread_create(&task1, NULL, do_task1, (void *)task1args) ) {
    fprintf(stderr, "%s(%lu): unable to create task1\n", __func__, pthread_self() );
    return 1;
  }

  if( pthread_create(&task2, NULL, do_task1, NULL) ) {
    fprintf(stderr, "%s(%lu): unable to create task2\n", __func__, pthread_self() );
    return 1;
  }

  printf("%d(%u)main: starting\n", getpid(), pthread_self() );
  sleep(10);
  for( counter = 20; counter; counter--) {
    task1_core();
    printf("%s(%u): tlsData=%u\n", __func__, pthread_self(), tlsData );
    sleep(1);
  }
  runtask1 = 0;
  for( counter = 2; !task1ended && counter; counter-- );
  return 0;
}
